grammar GroundAtom;

options
{
	output = AST;
	ASTLabelType = CommonTree;
	superClass = ExtendedParser; // contains all necessary code for proper Parser error handling
}

tokens
{
	// "Imaginary Tokens" fuer Tree-Labels
	PREDICATE;
	ARITY;
	RANGE_VALUE;
	ARGUMENTS;
	VARIABLE;
	CONSTANT;
	ATOM;
	
	QUERIES;
	QUERY;
	QUESTION_PART;
	GROUND_ATOM_EXPRESSIONS;
	ATOM_EXPRESSION;
	TRUE_ATOM_EXPRESSION;
	FALSE_ATOM_EXPRESSION;
	EMPTY;
}


@lexer::header
{
	package edu.cs.ai.kreator.parser.groundatom;
	import org.apache.log4j.Logger;
	import edu.cs.ai.kreator.exceptions.ParsingErrorInfo;
	import edu.cs.ai.kreator.exceptions.runtime.ParsingRuntimeException;
}

@parser::header
{
	package edu.cs.ai.kreator.parser.groundatom;
	import edu.cs.ai.kreator.parser.ExtendedParser;
}


// overriding these lexer methods prevents any (unwanted) automatic error recovery
@lexer::members
{
	// ********** Log4j Logger of this class **********
	public static final Logger LOG = Logger.getLogger(GroundAtomLexer.class);
    	// ************************************************

    	
	// This method is overridden to throw an exception (instead of just reporting the error)
	public void reportError(RecognitionException re) {
		throw new ParsingRuntimeException(re, getErrorMessage(re, getTokenNames()), ParsingErrorInfo.ErrorType.LEXICAL);
	}
    		
	// This method is overriden to send error messages to the LOG (instead of system.err)
	public void emitErrorMessage(String msg) {
		//System.err.println(msg);
		LOG.info(msg);
	}
}






//====================
//=== Parser Rules ===
//====================


// $<Start-Symbol

queriesInput
	:	NEWLINE*
		query
		(NEWLINE+ query)*
		NEWLINE*
		EOF
		-> ^(QUERIES query+)
	;

groundAtomsInput	
	:	groundAtomsFromFile
	;

// $>



// $<Query

query	:	'?-'? questionPart ('|' evidencePartFromQuery)?
		-> ^(QUERY questionPart evidencePartFromQuery?)
	;

questionPart
	:	atomExpression (',' atomExpression)*
		-> ^(QUESTION_PART atomExpression+)
	;

evidencePartFromQuery
	:	atomExpression (',' atomExpression)*
		-> ^(GROUND_ATOM_EXPRESSIONS atomExpression+)
	;

groundAtomsFromFile
	:	NEWLINE* EOF // empty evidence file
		-> ^(GROUND_ATOM_EXPRESSIONS EMPTY)
	|	NEWLINE* atomExpression ((',' NEWLINE*|';' NEWLINE*|NEWLINE+) atomExpression)* (',' NEWLINE*|';' NEWLINE*|NEWLINE+)? EOF
		-> ^(GROUND_ATOM_EXPRESSIONS atomExpression+)
	;

atomExpression
	:	atom '=' rangeValue
		-> ^(ATOM_EXPRESSION atom rangeValue)
	|	atom
		-> ^(TRUE_ATOM_EXPRESSION atom)
	|	'!' atom
		-> ^(FALSE_ATOM_EXPRESSION atom)
	;

// $>



// $<Logic Elements

atom	:	predicate arguments?
		-> ^(ATOM predicate arguments?)
	;

predicate
	:	identifierL
		-> ^(PREDICATE identifierL)
	;

arguments
	:	'(' term (',' term)* ')'
		-> ^(ARGUMENTS term+)
	;

term	:	variable | constant
	;


arity	:	DIGITS
		-> ^(ARITY DIGITS)
	;

rangeValue
	:	identifierL
		-> ^(RANGE_VALUE identifierL)
	;	

constant // query snytax: constant must be begin with a LOWER CASE character 
	:	identifierL
		-> ^(CONSTANT identifierL)
	;
	
variable // query snytax: variables must be begin with a UPPER CASE character
	:	identifierU
		-> ^(VARIABLE identifierU)
	;

// $>



// $<Identifiers and Keywords

identifierL
	:	IDENTIFIER_L
	;


identifierU
	:	IDENTIFIER_U
	;

// $>



//===================
//== Lexer Rules ===
//===================

// (vermutliche?) ANTLR Reihenfolge bei der Bestimmung des Token-Type:
// 1. explizit definierte Tokens mittels tokens{...}
// 2. Tokens, die mittels String-Literalen impliziet in Parser-Rules definiert sind
// 3. Tokens definiert durch Lexer Rules (in entsprechender Reihenfolge)
// Unterschied zwischen Prioritaet der "fragment" und der "normalen" Lexer-Rules???



// $<Fragments

fragment
IdentifierBody
	:	Alphanum | '_'
	;

fragment
Alphanum
	:	LetterUppercase | LetterLowercase | Digit
	;

fragment
LetterUppercase
	:	'A'..'Z'
	;

fragment
LetterLowercase
	:	'a'..'z'
	;

fragment
Digit
	:	'0'..'9'
	;

// $>
	
	
// $<Tokens

IDENTIFIER_L
	:	LetterLowercase IdentifierBody*
	;

IDENTIFIER_U
	:	LetterUppercase IdentifierBody*
	;

PROBABILITY	// PROBABILITY unterscheidet sich durch '.' eindeutig von einem "reinem" DIGITS.
		// Daher muss an entsprechender Stelle im allgemeinen POSITIVE_REAL und DIGITS erlaubt werden.
	:	('0' | '1') '.' DIGITS
	;

DIGITS
	:	Digit+
	;

NEWLINE
	:	'\r'? '\n'
	;	

// $>


// $<Skipped Tokens

WHITESPACE	
	:	(' ' | '\t' | '\f')+
		{$channel=HIDDEN;}
	;
		
LINE_COMMENT
    	: 	'//' ~('\n'|'\r')*
    		// zuvor benutzt:
    		// Zeielenkommentar am Ende der Datei ohne Newline liefert Fehler
    		// '//' ~('\n'|'\r')* '\r'? '\n'
	    	{$channel=HIDDEN;}
	;
	
BLOCK_COMMENT
    	:   	'/*' ( options {greedy=false;} : . )* '*/' {$channel=HIDDEN;}
    	;


// $>

