tree grammar GroundAtomWithoutLanguageWalker;


options
{
	tokenVocab = GroundAtom;
	ASTLabelType = CommonTree;
}

@header
{
	package edu.cs.ai.kreator.parser.groundatom;

	import org.apache.log4j.Logger;

	import edu.cs.ai.kreator.exceptions.*;
	import edu.cs.ai.kreator.exceptions.runtime.*;
	import edu.cs.ai.kreator.logic.*;
	import edu.cs.ai.kreator.logic.LanguageFactory.*;
	import edu.cs.ai.kreator.models.*;

	import edu.cs.ai.kreator.logic.Literal.*;

	import java.util.TreeSet;
	import java.util.Hashtable;


	/*	Werden schon von ANTLR importiert:
	import java.util.List;
	import java.util.ArrayList;
	import java.util.HashMap;
	import java.util.Map;
	*/
}

@members
{
	// ********** Log4j Logger of this class **********
	public static final Logger LOG = Logger.getLogger(GroundAtomWithoutLanguageWalker.class);
    	// ************************************************


	private boolean initialized = false;
	private LanguageFactory lang = null;
	private TreeSet<String> logicalRange;

	private TreeSet<String> predicateNames;
	private TreeSet<String> constantNames;
	private Hashtable<String, Integer> predicateArityMap;
	private Hashtable<String, TreeSet<String>> predicateRangeMap;

    	/**
    	 * This Method MUST be called before starting the walker with the method <code>groundAtomsInput</code>.
    	 */
    	public void init()
    	{
    		lang = new LanguageFactory();
    		logicalRange = new TreeSet<String>();
    		logicalRange.addAll(lang.getLogicalRange());
    		
    	
		predicateNames = new TreeSet<String>();
		constantNames = new TreeSet<String>();
		predicateArityMap = new Hashtable<String, Integer>();
		predicateRangeMap = new Hashtable<String, TreeSet<String>>();    	
    	
    		initialized = true;
    	}
    	
    	public LanguageFactory getLanguageFactory()
    	{
    		return lang;
    	}
    	
       	public TreeSet<String> getPredicateNames()
       	{
		return predicateNames;
	}
	
       	public TreeSet<String> getConstantNames()
       	{
		return constantNames;
	}
       	
       	public Hashtable<String, Integer> getPredicateArityMap()
       	{
       		return predicateArityMap;
       	}
       	
       	public Hashtable<String, TreeSet<String>> getPredicateRangeMap()
       	{
		return predicateRangeMap;
	}
       	
    	

}



// $<Sample


groundAtomsInput
	@init
	{
		if (initialized == false)
		{
			throw new RuntimeException("The method 'init' must be called before starting the walker.");
		}
	}
	:	groundAtomExpressions
		{
			// construct the LanguageFactory
			try
			{
				String output = "Predicates:\n" ;
				// add each predicate (with its determined arity and range) to the LanguageFactory
				for (String predName : predicateNames)
				{
					output += predName + "/" + predicateArityMap.get(predName) + " " + predicateRangeMap.get(predName) + "\n";
					
					// Code before logic rework:
					//lang.getPredicate(predName, predicateArityMap.get(predName), null, new ArrayList<String>(predicateRangeMap.get(predName)));
					
					// modified code: the predicate used to build the ground-atom is created on demand
					// (because predicates are defined implicitly)
					GeneralizedPredicate.createOrGet(lang, predName, predicateArityMap.get(predName), null, new ArrayList<String>(predicateRangeMap.get(predName)));
				}

				// add each constant to the LanguageFactory 
				for (String constName : constantNames)
				{			
					Constant.createOrGet(lang, constName);  
				}
				
				output += "Constants:\n" + constantNames;
				LOG.trace("Extracted from samples file:\n" + output);
			}
			catch (IllegalArgumentException ex)
			{
				throw new RuntimeException(ex);
			}
			catch (WriteProtectionException ex)
			{
				throw new RuntimeException(ex);
			}
            			
		}
	;


groundAtomExpressions
	:	^(GROUND_ATOM_EXPRESSIONS EMPTY) // empty file
	|	^(GROUND_ATOM_EXPRESSIONS atomExpression+)
	;

atomExpression
	:	^(ATOM_EXPRESSION atom rangeValue)
		{
			String predName  = $atom.value;
			String rangeName = $rangeValue.value;
			// if the predicate has no range set assigned yet, create one and assign range value
			if (!predicateRangeMap.containsKey(predName))
			{
				// if range value is logical, assign logical range
				if (rangeName.equals(lang.getLogicalRange().get(0))
					|| rangeName.equals(lang.getLogicalRange().get(1)))
				{
					predicateRangeMap.put(predName, logicalRange);
				}
				else
				{					
					predicateRangeMap.put(predName, new TreeSet<String>());
					// add the range value
					predicateRangeMap.get(predName).add(rangeName);					
				}
			}
			// check if predicate has logical range and range value is non-logical
			else if (predicateRangeMap.get(predName) == logicalRange
					&& !rangeName.equals(lang.getLogicalRange().get(0))
					&& !rangeName.equals(lang.getLogicalRange().get(1))
				)
			{		 
				throw new ParsingRuntimeException($atom.line, $atom.pos,
				"The predicate '" + predName + "' has the range value '" + rangeName + "', "
				+ " although it has logical range.",
				ParsingErrorInfo.ErrorType.SEMANTIC);			
			}
			else // add the range value a the existing range set
			{
				predicateRangeMap.get(predName).add(rangeName);
			}
			
		}
	|	(^(TRUE_ATOM_EXPRESSION a=atom)	| ^(FALSE_ATOM_EXPRESSION a=atom))
		{		
			String predName  = $a.value;
			// if the predicate has no range set assigned yet, assign logical Range
			if (!predicateRangeMap.containsKey(predName))
			{
				predicateRangeMap.put(predName, logicalRange);
			}
			// check if predicate has other range
			else if (predicateRangeMap.get(predName) != logicalRange)
			{		 
				throw new ParsingRuntimeException($a.line, $a.pos, 
				"The range of the predicate '" + predName + "' seems to be logical, "
				+ " but already contains the range values " + predicateRangeMap.get(predName),
				ParsingErrorInfo.ErrorType.SEMANTIC);			
			}
		}
	;

// $>



// $<Logic Elements
atom		returns[String value, int line, int pos]
	:	^(ATOM predicate arguments[$predicate.value,$predicate.line,$predicate.pos]?)
		{
			$value = $predicate.value;
			$line = $predicate.line;
			$pos = $predicate.pos;
			
			// if the predicate had no arguments, add 0 arity to the map
			if (!predicateArityMap.containsKey($predicate.value))
			{
				predicateArityMap.put($predicate.value, 0);
			}			
		}
	;

predicate	returns[String value, int line, int pos]
	:	^(PREDICATE identifierL)
		{	
			$value = $identifierL.value;
			$line  = $identifierL.line;
			$pos   = $identifierL.pos;
			
			predicateNames.add($value);
		}
	;

arguments[String predName, int line, int pos]	
	@init
	{
		int arity = 0;
	}
	:	^(ARGUMENTS (term
				  {
				  	arity++;
				  }
			    )+
		)
		{
			if (predicateArityMap.containsKey(predName))
			{
				if (predicateArityMap.get(predName).intValue() != arity)
				{
					throw new ParsingRuntimeException(line, pos,
					"The predicate '" + predName + "' has " + arity 
					+ " arguments, but has already been read with " + predicateArityMap.get(predName).intValue() + " arguments.",
					ParsingErrorInfo.ErrorType.SEMANTIC);
				}
			}
			else
			{
				predicateArityMap.put(predName, arity);
			}
		}
	;

term	:	constant
	;

constant:	^(CONSTANT identifierL)
		{
			String constantString = $identifierL.value;
			constantNames.add(constantString);
		}
	;
	
rangeValue	returns[String value]
	:	^(RANGE_VALUE identifierL)
		{
			$value = new String($identifierL.value);
		}
	;

// $>



// $<Identifiers and Keywords

identifierL returns [String value, int line, int pos]
	:	str=IDENTIFIER_L
		{
			$value = new String($str.text);
			$line  = $str.line;
			$pos   = $str.pos;
		}
	;


identifierU returns [String value, int line, int pos]
	:	str=IDENTIFIER_U
		{
			$value = new String($str.text);
			$line = $str.line;
			$pos = $str.pos;
		}
	;

// $>


